import os
import re
import sys
import subprocess
import platform
import argparse

# Consts
CHIP_LIST = ['mc30sf6', 'mct04', 'nvcom05', 'mct06']
BOARD_LIST = ['mc30sf6em', 'mct04em', 'nvcom05ipkyddr2', 'mct06em']
DEFAULT_TRIPLE = "mipsel-elf32-"

FLASH_PAGE_SIZE = 256
FLASH_SECTOR_SIZE = 64*1024
SPI_BOOT_SIZE = 256

def setupArgumentParser(chip_list, board_list):
	parser = argparse.ArgumentParser(description='Create binary bootloader for ELVEES Mips Multicores')

	# required argument
	parser.add_argument('-c', '--chip-name',
		choices=chip_list,
		metavar='chip', required=True,
		help="Name of the chip, the bootloader is building for")

	# one of next to arguments is required
	group = parser.add_mutually_exclusive_group(required=True)
	group.add_argument('-b','--board-name',
		choices=board_list,
		metavar='board',
		help="Name of the board, in case of using our memory initialization")
	group.add_argument('-i', '--ram-init',
		metavar='<ddr3init.elf>',
		help="Path to memory initialization elf file, in case of using external one")

	# required argument
	parser.add_argument('-f', '--firmware',
		required=True,
		metavar='<u-boot>',
		help="Path to firmware elf file")
	# optional argument
	parser.add_argument('-t', '--triple',
		required=False,
		metavar='<mipsel-elf32->',
		help="Triple to use as cross compile")

	return parser

def cleanPreviousBuild():
	subprocess.run(["rm", "-rf", "*.dis"],
		shell=(platform.system() == 'Windows'), check=True)
	subprocess.run(["rm", "-rf", "*.bin"],
		shell=(platform.system() == 'Windows'), check=True)
	subprocess.run(["rm", "-rf", "*.elf"],
		shell=(platform.system() == 'Windows'), check=True)
	subprocess.run(["rm", "-rf", "boards-memory-configs/build*"],
		shell=(platform.system() == 'Windows'), check=True)

def parseElfFile(triple, elf_file, only_text):
	# base_mem address
	base_mem_ret = subprocess.run(
		[triple + "readelf", "-S", "{}".format(elf_file)],
		check=True, stdout=subprocess.PIPE)
	base_mem_match = re.search('\[\s1\]\s\.\w+\s\s+\w+\s+([a-f0-9]+)',
		base_mem_ret.stdout.decode(),
		flags=re.DOTALL)
	base_mem = int(base_mem_match.group(1), base=16)

	# firmware entry point
	entry_point_ret = subprocess.run(
		[triple + "readelf", "-h", "{}".format(elf_file)],
		check=True, stdout=subprocess.PIPE)
	entry_point_match = re.search('Entry\spoint\saddress\:\s+0x([a-f0-9]+)',
		entry_point_ret.stdout.decode(), flags=re.DOTALL)
	entry_point = int(entry_point_match.group(1), base=16)

	# firmware binary size
	args = [triple + "objcopy", "-O", "binary",
			"{}".format(elf_file),
			"{}.bin".format(os.path.basename(elf_file))]
	if only_text:
		args.append("--only-section=.text")
	subprocess.run(args, check=True)
	size = os.stat("{}.bin".format(os.path.basename(elf_file))).st_size

	return [base_mem, entry_point, size]

def getSymbolAddress(triple, elf_file, symbol):
	ret = subprocess.run([triple + "readelf", "-s", "{}".format(elf_file)],
		check=True, stdout=subprocess.PIPE)
	match = re.search(
		'([a-f0-9]+)\s+\d+\s\w+\s+\w+\s\w+\s+\d+\s' + symbol,
		ret.stdout.decode(), flags=re.DOTALL)
	address = int(match.group(1), base=16)

	return address

def updateSpiBootXlFile(base):
	# build spi-boot1
	xl_file = open("boot-spi-base-address.xl", "w")
	xl_file.write(". = 0x{:x};".format(base))
	xl_file.close()

def compileSpiBoot(triple, name,
	firm_base, firm_entrypoint, firm_size, firm_base_bin,
	target, stage2):

	subprocess.run([triple + "gcc",
		"{}.S".format(name),
		"-DBASE_MEM=0x{:x}".format(firm_base),
		"-DFIRM_ENTRY=0x{:x}".format(firm_entrypoint),
		"-DFIRM_SIZE=0x{:x}".format(firm_size),
		"-DFIRM_BASE=0x{:x}".format(firm_base_bin),
		"-DTARGET_{}".format(target.upper()),
		"-DSPI_BOOT{}".format("2" if stage2 else "1"),
		"-Wa,--mno-fix-mc24r2",
		"-mips32", "-EL", "-nostdlib", "-nostartfiles", "-fno-delayed-branch", "-G0", "-O0",
		"-T{}.xl".format(name),
		"-o", "{}.elf".format(name + ("2" if stage2 else "1"))],
		check=True)

	dis = subprocess.run(
		[triple + "objdump", "-D", "{}1.elf".format(name)],
		check=True, stdout=subprocess.PIPE)
	dis_file = open("{}.dis".format(name + ("2" if stage2 else "1")), "w")
	dis_file.write(dis.stdout.decode())
	dis_file.close()
	subprocess.run([triple + "objcopy",
		"-O", "binary",
		"--only-section=.text",
		"{}.elf".format(name + ("2" if stage2 else "1")),
		"{}.bin".format(name + ("2" if stage2 else "1"))],
		check=True)


if __name__ == "__main__":
	parser = setupArgumentParser(CHIP_LIST, BOARD_LIST)
	args = parser.parse_args()

	board_name	= args.board_name
	chip_name	= args.chip_name
	firmware	= args.firmware
	ram_init	= args.ram_init
	triple		= args.triple if args.triple else DEFAULT_TRIPLE

	cc = triple + "gcc"
	objdump = triple + "objdump"
	objcopy = triple + "objcopy"
	readelf = triple + "readelf"
	nvcom05_firmware_name = "boot-spi"

	firmware_elf_filename = firmware
	firmware_elf_basename = os.path.basename(firmware_elf_filename)

	cleanPreviousBuild()

	# Set meminit to use
	if ram_init:
		print("Use given memory init elf file")
		meminit_elf_filename = ram_init
		meminit_elf_basename = os.path.basename(meminit_elf_filename)
	elif board_name:
		print("Building memory init elf file based on the board name")
		subprocess.run(["make", "BOARD_NAME={}".format(board_name), "-C", "boards-memory-configs", "TRIPLE={}".format(triple)], check=True, stdout=subprocess.PIPE)
		meminit_elf_filename = "boards-memory-configs/build/memory-config-{}.elf".format(board_name.lower())
		meminit_elf_basename = os.path.basename(meminit_elf_filename)
	else:
		assert(0) # this cant happen

	[meminit_base, meminit_entrypoint, meminit_size] = \
		parseElfFile(triple, meminit_elf_filename, 1 if board_name else 0)

	jump = getSymbolAddress(triple, meminit_elf_filename, "_spi_boot_2_start")
	print("_spi_boot_2_start: 0x{:x}".format(jump))

	# spi-boot1 must load ddrinit+spi_boot2, eval total size to load
	total_spi_boot1_load_size = jump - meminit_base + SPI_BOOT_SIZE
	print("------------- spi-boot-1 build defines -------------")
	print("BASE_MEM: 0x{:x}".format(meminit_base))
	print("FIRM_ENTRY: 0x{:x}".format(meminit_entrypoint))
	print("FIRM_SIZE: 0x{:x}".format(total_spi_boot1_load_size))
	print("FIRM_BASE: 0x{:x}".format(SPI_BOOT_SIZE))
	print("------------- Flash program offsets ----------------")
	print("spi-boot-1 program offset:\t0x{:x}".format(0))
	print("ddrinit program offset:\t\t0x{:x}".format(SPI_BOOT_SIZE))
	print("spi-boot-2 program offset:\t0x{:x}".format(
		SPI_BOOT_SIZE + jump - meminit_base))
	print("firmware program offset:\t0x{:x}".format(
		SPI_BOOT_SIZE + jump - meminit_base + SPI_BOOT_SIZE))

	updateSpiBootXlFile(0xb8000000)
	compileSpiBoot(triple, "boot-spi", meminit_base,
		meminit_entrypoint, total_spi_boot1_load_size,
		SPI_BOOT_SIZE, chip_name, 0)

	[firmware_base, firmware_entrypoint, firmware_size] = \
		parseElfFile(triple, firmware_elf_filename, 0)
	# eval firmware program offset
	firm_base_bin = SPI_BOOT_SIZE + jump - meminit_base + SPI_BOOT_SIZE

	print("------------- spi-boot-2 build defines -------------")
	print("BASE_MEM: 0x{:x}".format(firmware_base))
	print("FIRM_ENTRY: 0x{:x}".format(firmware_entrypoint))
	print("FIRM_SIZE: 0x{:x}".format(firmware_size))
	print("FIRM_BASE: 0x{:x}".format(firm_base_bin))

	updateSpiBootXlFile(jump)
	compileSpiBoot(triple, "boot-spi", firmware_base,
		firmware_entrypoint, firmware_size,
		firm_base_bin, chip_name, 1)

	image_boot_file	 = open("image-boot.img", "wb")
	boot_spi1_file	 = open("boot-spi1.bin", "rb")
	boot_spi2_file	 = open("boot-spi2.bin", "rb")
	meminit_bin_file = open("{}.bin".format(meminit_elf_basename), "rb")
	firm_bin_file	 = open("{}.bin".format(firmware_elf_basename), "rb")
	
	wrote_len = image_boot_file.write(boot_spi1_file.read(SPI_BOOT_SIZE))
	if (wrote_len < SPI_BOOT_SIZE):
		dummy_data = b'\xff' * (SPI_BOOT_SIZE - wrote_len)
		image_boot_file.write(dummy_data)

	image_boot_file.write(meminit_bin_file.read())
	dummy_bytes = jump - meminit_base - meminit_size
	if dummy_bytes > 0:
		dummy_data = b'\xff' * dummy_bytes
		image_boot_file.write(dummy_data)

	wrote_len = image_boot_file.write(boot_spi2_file.read(SPI_BOOT_SIZE))
	if (wrote_len < SPI_BOOT_SIZE):
		dummy_data = b'\xff' * (SPI_BOOT_SIZE - wrote_len)
		image_boot_file.write(dummy_data)

	image_boot_file.write(firm_bin_file.read())

	image_boot_file.close()
	boot_spi1_file.close()
	boot_spi2_file.close()
	meminit_bin_file.close()
	firm_bin_file.close()